/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Fibre Channel Host Bus Adapters.                                *
 * Copyright (C) 2003-2006 Emulex.  All rights reserved.           *
 * EMULEX and SLI are trademarks of Emulex.                        *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of version 2 of the GNU General       *
 * Public License as published by the Free Software Foundation.    *
 * This program is distributed in the hope that it will be useful. *
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND          *
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE      *
 * DISCLAIMED, EXCEPT TO THE EXTENT THAT SUCH DISCLAIMERS ARE HELD *
 * TO BE LEGALLY INVALID.  See the GNU General Public License for  *
 * more details, a copy of which can be found in the file COPYING  *
 * included with this package.                                     *
 *******************************************************************/

/* 
 * $Id: mpl_init.c 1018 2006-02-24 17:08:46Z jinfante $
 */

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/blk.h>
#include <linux/utsname.h>
#include <linux/list.h>

/* From drivers/scsi */
#include <sd.h>
#include <hosts.h>
#include <scsi.h>
#else
#include <linux/spinlock.h>
#include <linux/pci.h>
#include <linux/blkdev.h>
#include <linux/moduleparam.h>
#include <scsi/scsi.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_transport_fc.h>
#endif

#include <linux/ctype.h>
#include <linux/string.h>

#include "mpl.h"
#include "mpl_mem.h"
#include "mpl_log.h"
#include "mpl_crtn.h"

#define MPL_SYNTAX_ERR   0xff	

#define IGNORE_LPFC_CONFIGURATION
#define MPULSE

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
char *lpfc_multipulse[MPL_MAX_STATIC_DEVICES] = { 0 };
char *mpl_exclusion[MPL_MAX_STATIC_DEVICES] = { 0 };
int  mpl_max_static_devices = MPL_MAX_STATIC_DEVICES;
int  mpl_max_exclusion_devices = MPL_MAX_EXCLUSION_DEVICES;
module_param_array(lpfc_multipulse, charp, mpl_max_static_devices, 0);
module_param_array(mpl_exclusion, charp, mpl_max_exclusion_devices, 0);

int  mpl_use_static_config = 1;
int  mpl_use_exclusion_list = 0;
module_param(mpl_use_static_config, int, 0);
module_param(mpl_use_exclusion_list, int, 0);
#else
#include "/etc/lpfc.conf"
MODULE_PARM(lpfc_multipulse, "1-" __MODULE_STRING(MPL_MAX_STATIC_DEVICES) "s");
MODULE_PARM(mpl_exclusion, "1-" __MODULE_STRING(MPL_MAX_STATIC_DEVICES) "s");
MODULE_PARM(mpl_use_static_config, "i");
MODULE_PARM(mpl_use_exclusion_list, "i");
#endif
MODULE_PARM_DESC(lpfc_multipulse, "Define static MultiPulse configuration");
MODULE_PARM_DESC(mpl_exclusion, "Define an exclusion list of Paths");
MODULE_PARM_DESC(mpl_use_static_config, "Use static configuration entries");
MODULE_PARM_DESC(mpl_use_exclusion_list, "Use exclusion list entries");

struct config_device {
	struct config_device *next;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	int  host_no;
	int  target_no;
	int  lun_no;
#else
	size_t   serial_size;	      /* actual bytes in serial number */
	uint8_t  serial_number[256];
#endif
	int  flags;
};

struct config_path {
	struct config_path *next;
	struct config_device *device;
	uint8_t  hba_wwpn[8];
	uint8_t  target_wwpn[8];
	int      lun_no;
	int      flags;
};

static void *config_space;
static void *config_ptr;
static struct config_device *mpl_config_devices;
static struct config_path *mpl_config_paths;


int mpl_get_path_flags(int *flags, uint8_t host_wwpn[8],
		       uint8_t target_wwpn[8], int lun) 
{
	struct config_path *path;

	for (path = mpl_config_paths; path; path = path->next) {
		if (!memcmp(path->hba_wwpn, host_wwpn, sizeof(path->hba_wwpn)) &&
		    !memcmp(path->target_wwpn, target_wwpn, sizeof(path->target_wwpn)) &&
		    path->lun_no == lun) {
			*flags = path->flags;
			return 0;
		}
	}
	return -1;
}

int mpl_get_device_flags_by_path(int *flags, uint8_t host_wwpn[8],
				 uint8_t target_wwpn[8], int lun) 
{
	struct config_path *path;

	for (path = mpl_config_paths; path; path = path->next) {
		if (!memcmp(path->hba_wwpn, host_wwpn, sizeof(path->hba_wwpn)) &&
		    !memcmp(path->target_wwpn, target_wwpn, sizeof(path->target_wwpn)) &&
		    path->lun_no == lun) {
			*flags = path->device->flags;
			return 0;
		}
	}
	return -1;
}



static int strtoi(const char *cp, char **endp, int base)
{
	while (isspace(*cp))
		++cp;
	if (*cp == '-')
		return -simple_strtoul(cp + 1, endp, 0);
	else
		return simple_strtoul(cp, endp, 0);
}

static inline void
mpl_crack_wwn_no_colon(unsigned char wwn[8], const char *cp,
		       const char **wwn_buffer)
{
	int  i;

	memset(wwn, 0, 8);
	for (i = 0; i < 8; ++i) {
		while (isspace(*cp)) {
			++cp;
		}
		if (!isxdigit(*cp)) {
			break;
		}
		wwn[i] = mpl_xtobin(*cp);
		++cp;
		if (isxdigit(*cp)) {
			wwn[i] = (wwn[i] << 4) + mpl_xtobin(*cp);
			++cp;
		}
	}
	if (wwn_buffer) {
		*wwn_buffer = cp;
	}
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
static inline int
mpl_crack_serial(unsigned char *serial_number, int serial_size, const char *cp,
		 const char **next_cp)
{
	unsigned char *sb = serial_number;
	int  i;

	memset(sb, 0, serial_size);
	for (i = 0; i < serial_size; ++i) {
		while (isspace(*cp)) {
			++cp;
		}
		if (!isxdigit(*cp)) {
			break;
		}
		*sb = mpl_xtobin(*cp);
		++cp;
		if (isxdigit(*cp)) {
			*sb = (*sb << 4) + mpl_xtobin(*cp);
			++sb;
			++cp;
		}
	}
	if (next_cp) {
		*next_cp = cp;
	}
	return sb - serial_number;
}
#endif

static struct config_device *
mpl_config_device(const char *s, char **ptr)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	struct config_device *device;
	int  hba, target, lun, flags;
	const char *cp = s;
	char *endp;

	while (isspace(*cp))
		++cp;
	if (memcmp(cp, "lpfc", 4)) {
		mpl_print(0, MPL_ERR "Bad device specification: %s\n", s);
		return NULL;
	}
	hba = strtoi(cp + 4, &endp, 0);
	cp = endp;
	if (*cp != 't') {
		mpl_print(0, MPL_ERR "Bad hba number: %s\n", s);
		return NULL;
	}
	target = strtoi(cp + 1, &endp, 0);
	cp = endp;
	if (*cp != 'd') {
		mpl_print(0, MPL_ERR "Bad target number: %s\n", s);
		return NULL;
	}
	lun = strtoi(cp + 1, &endp, 0);
	cp = endp;
	if (*cp != ':') {
		mpl_print(0, MPL_ERR "Bad lun number: %s\n", s);
		return NULL;
	}
	flags = strtoi(cp + 1, &endp, 0);
	cp = endp;
	while (isspace(*cp) || *cp == ',' || *cp == '"' || *cp == '\'') {
		++cp; ++endp;
	}
	if (*cp != 0) {
		*ptr = endp;
	}
	else {
		*ptr = NULL;
	}

	device = config_ptr;
	config_ptr = device + 1;
	device->host_no = hba;
	device->target_no = target;
	device->lun_no = lun;
	device->flags = flags;
	device->next = mpl_config_devices;
	mpl_config_devices = device;

	return device;
#else
	struct config_device *device;
	int  serial_size;
	unsigned char serial_number[256];
	int  flags;
	const char *cp = s;
	char *endp;

	while (isspace(*cp))
		++cp;
	serial_size = mpl_crack_serial(serial_number, sizeof(serial_number),
				       cp, &cp);
	if (*cp != ':') {
		mpl_print(0, MPL_ERR "Bad device serial: %s (%s)\n", s, cp);
		return NULL;
	}
	flags = strtoi(cp + 1, &endp, 0);
	cp = endp;
	while (isspace(*cp) || *cp == ',' || *cp == '"' || *cp == '\'') {
		++cp; ++endp;
	}
	if (*cp != 0) {
		*ptr = endp;
	}
	else {
		*ptr = NULL;
	}
	device = config_ptr;
	config_ptr = device + 1;
	memcpy(device->serial_number, serial_number,
	       sizeof(device->serial_number));
	device->serial_size = serial_size;
	device->flags = flags;
	device->next = mpl_config_devices;
	mpl_config_devices = device;
	return device;
#endif
}

static struct config_path *
mpl_config_path(struct config_device *device, const char *s, char **ptr)
{
	struct config_path *path;
	unsigned char host_wwpn[8];
	unsigned char target_wwpn[8];
	unsigned int  lun;
	int  flags;
	const char *cp = s;
	char *endp;

	mpl_crack_wwn_no_colon(host_wwpn, cp, &cp);
	if (memcmp(cp, ":tgt:", 5)) {
		mpl_print(0, MPL_ERR "Bad hba wwpn: %s\n", s);
		return NULL;
	}
	mpl_crack_wwn_no_colon(target_wwpn, cp + 5, &cp);
	if (memcmp(cp, ":lun:", 5)) {
		mpl_print(0, MPL_ERR "Bad target wwpn: %s\n", s);
		return NULL;
	}
	lun = simple_strtoul(cp + 5, &endp, 0);	
	cp = endp;
	if (*cp != ':') {
		mpl_print(0, MPL_ERR "Bad lun number: %s\n", s);
		return NULL;
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	if (device->lun_no != lun) {
		mpl_print(0, MPL_ERR "lun mismatch, device %d, path %d: %s\n",
			  device->lun_no, lun, s);
	}
#endif
	flags = simple_strtoul(cp + 1, &endp, 0);
	cp = endp;
	while (isspace(*cp) || *cp == ',' || *cp == '"' || *cp == '\'') {
		++cp; ++endp;
	}
	if (*cp != 0) {
		*ptr = endp;
	}
	else {
		*ptr = NULL;
	}

	path = config_ptr;
	config_ptr = path + 1;

	memcpy(path->hba_wwpn, host_wwpn, sizeof(path->hba_wwpn));
	memcpy(path->target_wwpn, target_wwpn, sizeof(path->target_wwpn));
	path->lun_no = lun;
	path->flags = flags;
	path->device = device;
	path->next = mpl_config_paths;
	mpl_config_paths = path;

	return path;
}

static void
mpl_parse_entry(void)
{
	struct config_device *device = NULL;
	char *cp;
	char *ptr;
	int  i;

	if (!config_space)
		return;
	for (i = 0;
	     i < ARRAY_SIZE(lpfc_multipulse) && lpfc_multipulse[i] != 0;
	     ++i) {
		cp = lpfc_multipulse[i];
		while (isspace(*cp) || *cp == '"' || *cp == '\'')
			++cp;

		/* This while loop is just in case multiple entries are on one line */
		while (cp) {
			if (!memcmp(cp, "dev:", 4)) {
				device = mpl_config_device(cp + 4, &ptr);
				if (!device) {
					mpl_print(0, MPL_ERR
					  "Syntax error: invalid device %s\n",
					  lpfc_multipulse[i]);
				}
			} else if (!memcmp(cp, "hba:", 4)) {
				if (!device || !mpl_config_path(device, cp + 4, &ptr)) {
					mpl_print(0, MPL_ERR
					  "Syntax error: invalid path %s\n",
					  lpfc_multipulse[i]);
				}
			}
			cp = ptr;
		}
	}
}

static struct config_device *
mpl_dummy_device(void)
{
	struct config_device *device;

	device = config_ptr;
	config_ptr = device + 1;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
				/* Ensure we don't match serial number
				 * with null serial number.
				 */
	device->serial_size = sizeof(device->serial_size);
#endif
	device->flags = 0;
	device->next = mpl_config_devices;
	mpl_config_devices = device;
	return device;
}

static void
mpl_parse_exclusion(void)
{
	struct config_device *device = mpl_dummy_device();
	char *cp;
	char *ptr;
	int  i;

	if (!config_space)
		return;
	for (i = 0;
	     i < ARRAY_SIZE(mpl_exclusion) && mpl_exclusion[i] != 0;
	     ++i) {
		cp = mpl_exclusion[i];
		while (isspace(*cp) || *cp == '"' || *cp == '\'')
			++cp;

		/* This while loop is just in case multiple entries are on one line */
		while (cp) {
			if (!memcmp(cp, "hba:", 4)) {
				if (!device || !mpl_config_path(device, cp + 4, &ptr)) {
					mpl_print(0, MPL_ERR
					  "Syntax error: invalid path %s\n",
					  mpl_exclusion[i]);
				} else {
					mpl_config_paths->flags |= MPL_CFG_PATH_EXPORT;
				}
			}
			cp = ptr;
		}
	}
}

static int
mpl_alloc_config_tables(void)
{
	int  device_count = 0;
	int  path_count = 0;
	char *cp;
	int  i;

	if (mpl_use_static_config) {
		for (i = 0;
		     i < ARRAY_SIZE(lpfc_multipulse) && lpfc_multipulse[i] != 0;
		     ++i) {
			cp = lpfc_multipulse[i];
			while (isspace(*cp) || *cp == '"' || *cp == '\'')
				++cp;
			if (!memcmp(cp, "dev:", 4)) {
				++device_count;
			} else if (!memcmp(cp, "hba:", 4)) {
				++path_count;
			} else {
				mpl_print(0, MPL_ERR "Syntax error: unrecognized "
				  "configuration: %s\n", lpfc_multipulse[i]);
				return 1;
			}
		}
	}
	if (mpl_use_exclusion_list) {
		if (mpl_exclusion[0] != 0) {
			++device_count;
		}
		for (i = 0;
		     i < ARRAY_SIZE(mpl_exclusion) && mpl_exclusion[i] != 0;
		     ++i) {
		
			cp = mpl_exclusion[i];
			while (isspace(*cp) || *cp == '"' || *cp == '\'')
				++cp;
			if (!memcmp(cp, "hba:", 4)) {
				++path_count;
			} else {
				mpl_print(0, MPL_ERR "Syntax error: unrecognized "
				  "exclusion: %s\n", mpl_exclusion[i]);
				return 1;
			}
		}
	}
	if (device_count && path_count) {
		int size = (device_count * sizeof(struct config_device) +
			    path_count * sizeof(struct config_path));

		config_space = kmalloc(size, GFP_KERNEL);
		config_ptr = config_space;
		if (config_space) {
			memset(config_space, size, 0);
		}
	}
	return 0;
}

int
mpl_setup_multipulse(void)
{
	if ((ARRAY_SIZE(lpfc_multipulse) && lpfc_multipulse[0] != NULL) ||
	    (ARRAY_SIZE(mpl_exclusion) && mpl_exclusion[0] != NULL)) {
		if (mpl_alloc_config_tables())
			return 1;
	}

	if (mpl_use_static_config && ARRAY_SIZE(lpfc_multipulse) &&
						lpfc_multipulse[0] != NULL) {
		mpl_parse_entry();
	}
	if (mpl_use_exclusion_list && ARRAY_SIZE(mpl_exclusion) &&
						mpl_exclusion[0] != NULL) {
		mpl_parse_exclusion();
	}

	return 0;
}

void mpl_free_initial_data(void)
{
	kfree(config_space);
	config_space = NULL;
}
