/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Fibre Channel Host Bus Adapters.                                *
 * Copyright (C) 2003-2006 Emulex.  All rights reserved.           *
 * EMULEX and SLI are trademarks of Emulex.                        *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of version 2 of the GNU General       *
 * Public License as published by the Free Software Foundation.    *
 * This program is distributed in the hope that it will be useful. *
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND          *
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE      *
 * DISCLAIMED, EXCEPT TO THE EXTENT THAT SUCH DISCLAIMERS ARE HELD *
 * TO BE LEGALLY INVALID.  See the GNU General Public License for  *
 * more details, a copy of which can be found in the file COPYING  *
 * included with this package.                                     *
 *******************************************************************/

/* 
 * $Id: mpl_mem.c 1005 2006-01-27 17:27:22Z hdaschbach $
 */

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/blk.h>
#include <linux/utsname.h>
#include <linux/list.h>

/* From drivers/scsi */
#include <sd.h>
#include <hosts.h>
#include <scsi.h>
#else
#include <linux/spinlock.h>
#include <linux/pci.h>
#include <linux/blkdev.h>
#include <scsi/scsi.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_transport_fc.h>
#endif

#include <linux/ctype.h>
#include <linux/string.h>

#include "mpl.h"
#include "mpl_log.h"
#include "mpl_crtn.h"

extern kmem_cache_t *io_stub_cache;
kmem_cache_t *scsi_cmd_cache;
kmem_cache_t *data_buf_cache;

#define MAX_MEM_CNT      32
struct mpl_mem_pool {
	void *buf_array[MAX_MEM_CNT];	
	uint16_t   max_count;
	uint16_t   cur_count;
};

struct mpl_mem_pool cmd_list;
struct mpl_mem_pool buf_list;
struct mpl_mem_pool stub_list;
/* lock for cmd_list, buf_list and stub_list */
static spinlock_t int_cmd_data_lock = SPIN_LOCK_UNLOCKED;

int
mpl_create_mem_pool(void)
{
	void *buf;
	int i;
	int rc = 0;

	scsi_cmd_cache = kmem_cache_create("mpl_scsi_cmd",
					  sizeof(struct scsi_cmnd), 0,
					  SLAB_HWCACHE_ALIGN,
					  NULL, NULL);

	data_buf_cache = kmem_cache_create("mpl_data_buf",
					  512, 0,
					  SLAB_HWCACHE_ALIGN,
					  NULL, NULL);

	if ((scsi_cmd_cache == NULL) || (data_buf_cache == NULL)) {
		goto mem_alloc_error;
	}

	memset(&cmd_list, 0, sizeof(cmd_list));
	memset(&buf_list, 0, sizeof(buf_list));
	memset(&stub_list, 0, sizeof(stub_list));

	for (i = 0; i < MAX_MEM_CNT; i++) {
		buf = kmem_cache_alloc(scsi_cmd_cache, SLAB_KERNEL);
		if (buf == NULL) {
			break;
		} else {
			cmd_list.buf_array[i] = buf;
			cmd_list.max_count++;
			cmd_list.cur_count++;
		}
	}

	for (i = 0; i < MAX_MEM_CNT; i++) {
		buf = kmem_cache_alloc(data_buf_cache, SLAB_KERNEL);
		if (buf == NULL) {
			break;
		} else {
			buf_list.buf_array[i] = buf;
			buf_list.max_count++;
			buf_list.cur_count++;
		}
	}

	for (i = 0; i < MAX_MEM_CNT; i++) {
		buf = kmem_cache_alloc(io_stub_cache, SLAB_KERNEL);
		if (buf == NULL) {
			break;
		} else {
			stub_list.buf_array[i] = buf;
			stub_list.max_count++;
			stub_list.cur_count++;
		}
	}

	if ((cmd_list.cur_count == 0) || 
	    (buf_list.cur_count == 0) ||
	    (stub_list.cur_count == 0)) {
mem_alloc_error:
		/* one of the list is empty so free all and return error */
		rc = -ENOMEM;
		for (i = 0; i < cmd_list.cur_count; i++) {
			kmem_cache_free(scsi_cmd_cache, cmd_list.buf_array[i]);
		}
		cmd_list.cur_count = 0;

		for (i = 0; i < buf_list.cur_count; i++) {
			kmem_cache_free(data_buf_cache, buf_list.buf_array[i]);
		}
		buf_list.cur_count = 0;

		for (i = 0; i < stub_list.cur_count; i++) {
			kmem_cache_free(io_stub_cache, stub_list.buf_array[i]);
		}
		stub_list.cur_count = 0;
	}

	return rc;
}


void
mpl_destroy_mem_pool(void)
{
	unsigned long flags;
	int i;
	kmem_cache_t *tmp;

	spin_lock_irqsave(&int_cmd_data_lock, flags); 

	for (i = 0; i < cmd_list.cur_count; i++) {
		kmem_cache_free(scsi_cmd_cache, cmd_list.buf_array[i]);
	}
	cmd_list.cur_count = 0;

	for (i = 0; i < buf_list.cur_count; i++) {
		kmem_cache_free(data_buf_cache, buf_list.buf_array[i]);
	}
	buf_list.cur_count = 0;

	for (i = 0; i < stub_list.cur_count; i++) {
		kmem_cache_free(io_stub_cache, stub_list.buf_array[i]);
	}
	stub_list.cur_count = 0;

	if (scsi_cmd_cache) {
		tmp = scsi_cmd_cache;
		scsi_cmd_cache = NULL;
		spin_unlock_irqrestore(&int_cmd_data_lock, flags); 
		kmem_cache_destroy(tmp);
		spin_lock_irqsave(&int_cmd_data_lock, flags); 
	}
	if (data_buf_cache) {
		tmp = data_buf_cache;
		data_buf_cache = NULL;
		spin_unlock_irqrestore(&int_cmd_data_lock, flags); 
		kmem_cache_destroy(tmp);
		spin_lock_irqsave(&int_cmd_data_lock, flags); 
	}
	spin_unlock_irqrestore(&int_cmd_data_lock, flags); 
}


struct scsi_cmnd *
mpl_get_scsi_cmd(void)
{
	unsigned long flags;
	struct scsi_cmnd *cmd;

	cmd = kmem_cache_alloc(scsi_cmd_cache, SLAB_ATOMIC);
	if (cmd == NULL) {
		spin_lock_irqsave(&int_cmd_data_lock, flags);
		if(cmd_list.cur_count) {
			cmd_list.cur_count--;
			mpl_print(MPL_INIT, MPL_INFO "POOL: Getting cmd at "
				  "index %x\n",
				  cmd_list.cur_count);
			cmd = cmd_list.buf_array[cmd_list.cur_count];
		}
		spin_unlock_irqrestore(&int_cmd_data_lock, flags);
	}
	if (cmd) {		/* Distribute stubs over hash table. */
		static unsigned long serial_number;

		memset(cmd, 0, sizeof(*cmd));
		cmd->serial_number = ++serial_number;
		cmd->allowed = 5;
		init_timer(&cmd->eh_timeout);
		cmd->eh_timeout.function = mpl_cmd_timeout;
		cmd->eh_timeout.expires = jiffies + 3600 * HZ;
		cmd->eh_timeout.data = 0;
	}

	return cmd;
}


void
mpl_put_scsi_cmd(struct scsi_cmnd *cmd)
{
	unsigned long flags;

	spin_lock_irqsave(&int_cmd_data_lock, flags);
	if (cmd_list.cur_count < cmd_list.max_count) {
		mpl_print(MPL_INIT, MPL_INFO "POOL: Putting cmd at index "
			  "%x\n",
			  cmd_list.cur_count);
		cmd_list.buf_array[cmd_list.cur_count] = cmd;
		cmd_list.cur_count++;
		spin_unlock_irqrestore(&int_cmd_data_lock, flags);
	} else {
		spin_unlock_irqrestore(&int_cmd_data_lock, flags);
		kmem_cache_free(scsi_cmd_cache, cmd);
	}	
}


void *
mpl_get_scsi_buf(void)
{
	unsigned long flags;
	void *buf;	

	buf = kmem_cache_alloc(data_buf_cache, SLAB_ATOMIC);
	if (buf == NULL) {
		spin_lock_irqsave(&int_cmd_data_lock, flags);
		if (buf_list.cur_count) {
			buf_list.cur_count--;
			mpl_print(MPL_INIT, MPL_INFO "POOL: Getting buf at "
				  "index %x\n",
				  buf_list.cur_count);
			buf = buf_list.buf_array[buf_list.cur_count];
		}
		spin_unlock_irqrestore(&int_cmd_data_lock, flags);
	}

	return buf;
}


void
mpl_put_scsi_buf(void *buf)
{
	unsigned long flags;

	spin_lock_irqsave(&int_cmd_data_lock, flags);
	if (buf_list.cur_count < buf_list.max_count) {
		mpl_print(MPL_INIT, MPL_INFO "POOL: Putting buf at index "
			  "%x\n",
			  buf_list.cur_count);
		buf_list.buf_array[buf_list.cur_count] = buf;
		buf_list.cur_count++;
		spin_unlock_irqrestore(&int_cmd_data_lock, flags);
	} else {
		spin_unlock_irqrestore(&int_cmd_data_lock, flags);
		kmem_cache_free(data_buf_cache, buf);
	}	
}



static struct io_stub *
mpl_get_stub(void)
{
	unsigned long flags;
	struct io_stub *stub;

	stub = kmem_cache_alloc(io_stub_cache, SLAB_ATOMIC);
	if (stub == NULL) {
		spin_lock_irqsave(&int_cmd_data_lock, flags);
		if(stub_list.cur_count) {
			stub_list.cur_count--;
			mpl_print(MPL_INIT, MPL_INFO "POOL: Getting stub at "
				  "index %x\n",
				  stub_list.cur_count);
			stub = stub_list.buf_array[stub_list.cur_count];
		}
		spin_unlock_irqrestore(&int_cmd_data_lock, flags);
	}

	return stub;
}

static void
mpl_put_stub(struct io_stub *stub)
{
	unsigned long flags;

	spin_lock_irqsave(&int_cmd_data_lock, flags);
	if (stub_list.cur_count < stub_list.max_count) {
		mpl_print(MPL_INIT, MPL_INFO "POOL: Putting stub at index "
			  "%x\n",
			  stub_list.cur_count);
		stub_list.buf_array[stub_list.cur_count] = stub;
		stub_list.cur_count++;
		spin_unlock_irqrestore(&int_cmd_data_lock, flags);
	} else {
		spin_unlock_irqrestore(&int_cmd_data_lock, flags);
		kmem_cache_free(io_stub_cache, stub);
	}	
}


struct io_stub *
mpl_stub_alloc(struct scsi_cmnd *cmd, struct mpl_path *path,
               void (*done)(struct scsi_cmnd *))
{
	struct io_stub *stub;

	stub = mpl_get_stub(); 
	if (!stub) {
		return NULL;
	}

	memset(stub, 0, sizeof(*stub));
	wmb();
	atomic_inc(&path->device->ref_count);
	stub->cmd = cmd;
	stub->done = done;
	stub->path = path;
	stub->sdev = cmd->device;
	return stub;
}

void
mpl_stub_free(struct io_stub *stub)
{
	atomic_dec(&stub->path->device->ref_count);
	mpl_put_stub(stub);
}
